@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright 2008 Zuza Software Foundation
+# Copyright 2008-2009 Zuza Software Foundation
 #
 # This file is part of Virtaal.
 #
@@ -47,6 +47,7 @@ class SearchMode(BaseMode):
             @type  controller: virtaal.controllers.ModeController
             @param controller: The ModeController that managing program modes."""
         self.controller = controller
+        self.unitview = controller.main_controller.unit_controller.view
 
         self._create_widgets()
         self._setup_key_bindings()
@@ -65,14 +66,22 @@ class SearchMode(BaseMode):
         self.btn_search.connect('clicked', self._on_search_clicked)
         self.chk_casesensitive = gtk.CheckButton(_('_Case sensitive'))
         self.chk_casesensitive.connect('toggled', self._refresh_proxy)
+        # l10n: To read about what regular expressions are, see
+        # http://en.wikipedia.org/wiki/Regular_expression
         self.chk_regex = gtk.CheckButton(_("_Regular expression"))
         self.chk_regex.connect('toggled', self._refresh_proxy)
 
         # Widgets for replace (second row)
+        # l10n: This text label shows in front of the text box where the replacement
+        # text is typed. Keep in mind that the text box will appear after this text.
+        # If this sentence construction is hard to use, consdider translating this as
+        # "Replacement"
         self.lbl_replace = gtk.Label(_('Replace with'))
         self.ent_replace = gtk.Entry()
+        # l10n: Button text
         self.btn_replace = gtk.Button(_('Replace'))
         self.btn_replace.connect('clicked', self._on_replace_clicked)
+        # l10n: Check box
         self.chk_replace_all = gtk.CheckButton(_('Replace _All'))
 
         self.widgets = [
@@ -108,13 +117,16 @@ class SearchMode(BaseMode):
 
         self._add_widgets()
         self._connect_highlighting()
+        self._connect_textboxes()
         if not self.ent_search.get_text():
             self.storecursor.indices = self.storecursor.model.stats['total']
         else:
             self.update_search()
 
+        curpos = self.ent_search.props.cursor_position
         def grab_focus():
             self.ent_search.grab_focus()
+            self.ent_search.set_position(curpos)
             return False
 
         # FIXME: The following line is a VERY UGLY HACK, but at least it works.
@@ -126,31 +138,31 @@ class SearchMode(BaseMode):
         main_controller.select_unit(match.unit)
         view = main_controller.unit_controller.view
 
-        # Wait for SearchMode to finish with its highlighting and stuff, and then we do...
         if match.part == 'target':
-            def select_match_text():
-                target = view.targets[match.part_n]
-                target.grab_focus()
-                buff = target.get_buffer()
-                buffstr = buff.get_text(buff.get_start_iter(), buff.get_end_iter()).decode('utf-8')
-                unescaped = markup.unescape(buffstr)
-                start, end = self._escaped_indexes(unescaped, match.start, match.end)
-                start_iter = buff.get_iter_at_offset(start)
-                end_iter = buff.get_iter_at_offset(end)
-                buff.select_range(end_iter, start_iter)
-                return False
+            textbox = view.targets[match.part_n]
         elif match.part == 'source':
-            def select_match_text():
-                source = view.sources[match.part_n]
-                source.grab_focus()
-                buff = source.get_buffer()
-                buffstr = buff.get_text(buff.get_start_iter(), buff.get_end_iter()).decode('utf-8')
-                unescaped = markup.unescape(buffstr)
-                start, end = self._escaped_indexes(unescaped, match.start, match.end)
-                start_iter = buff.get_iter_at_offset(match.start)
-                end_iter = buff.get_iter_at_offset(match.end)
-                buff.select_range(end_iter, start_iter)
-                return False
+            textbox = view.sources[match.part_n]
+
+        if not textbox:
+            return False
+
+        # Wait for SearchMode to finish with its highlighting and stuff, and then we do...
+        def select_match_text():
+            textbox.grab_focus()
+            buff = textbox.buffer
+            buffstr = textbox.get_text()
+            unescaped = markup.unescape(buffstr)
+
+            start, end = self._escaped_indexes(unescaped, match.start, match.end)
+            if hasattr(textbox.elem, 'gui_info'):
+                start = textbox.elem.gui_info.tree_to_gui_index(start)
+                end = textbox.elem.gui_info.tree_to_gui_index(end)
+            start_iter = buff.get_iter_at_offset(start)
+            end_iter = buff.get_iter_at_offset(end)
+
+            buff.select_range(end_iter, start_iter)
+            return False
+
         # TODO: Implement for 'notes' and 'locations' parts
         gobject.idle_add(select_match_text)
 
@@ -215,17 +227,22 @@ class SearchMode(BaseMode):
             self.storecursor.indices = self.storecursor.model.stats['total']
         self._highlight_matches()
 
+        curpos = self.ent_search.props.cursor_position
         def grabfocus():
             self.ent_search.grab_focus()
-            self.ent_search.set_position(-1)
+            self.ent_search.set_position(curpos)
             return False
         gobject.idle_add(grabfocus)
 
     def unselected(self):
         # TODO: Unhightlight the previously selected unit
-        if getattr(self, '_signalid_cursor_changed', ''):
+        if hasattr(self, '_signalid_cursor_changed'):
             self.storecursor.disconnect(self._signalid_cursor_changed)
 
+        if hasattr(self, '_textbox_signals'):
+            for textbox, signal_id in self._textbox_signals.items():
+                textbox.disconnect(signal_id)
+
         if self._unit_modified_id:
             self.controller.main_controller.unit_controller.disconnect(self._unit_modified_id)
             self._unit_modified_id = 0
@@ -235,21 +252,29 @@ class SearchMode(BaseMode):
     def _add_widgets(self):
         table = self.controller.view.mode_box
 
-        table.attach(self.ent_search, 2, 3, 0, 1)
-        table.attach(self.btn_search, 3, 4, 0, 1)
-        table.attach(self.chk_casesensitive, 4, 5, 0, 1)
-        table.attach(self.chk_regex, 5, 6, 0, 1)
+        xoptions = gtk.FILL
+        table.attach(self.ent_search, 2, 3, 0, 1, xoptions=xoptions)
+        table.attach(self.btn_search, 3, 4, 0, 1, xoptions=xoptions)
+        table.attach(self.chk_casesensitive, 4, 5, 0, 1, xoptions=xoptions)
+        table.attach(self.chk_regex, 5, 6, 0, 1, xoptions=xoptions)
 
-        table.attach(self.lbl_replace, 1, 2, 1, 2)
-        table.attach(self.ent_replace, 2, 3, 1, 2)
-        table.attach(self.btn_replace, 3, 4, 1, 2)
-        table.attach(self.chk_replace_all, 4, 5, 1, 2)
+        table.attach(self.lbl_replace, 1, 2, 1, 2, xoptions=xoptions)
+        table.attach(self.ent_replace, 2, 3, 1, 2, xoptions=xoptions)
+        table.attach(self.btn_replace, 3, 4, 1, 2, xoptions=xoptions)
+        table.attach(self.chk_replace_all, 4, 5, 1, 2, xoptions=xoptions)
 
         table.show_all()
 
     def _connect_highlighting(self):
         self._signalid_cursor_changed = self.storecursor.connect('cursor-changed', self._on_cursor_changed)
 
+    def _connect_textboxes(self):
+        self._textbox_signals = {}
+        for textbox in self.unitview.sources + self.unitview.targets:
+            self._textbox_signals[textbox] = textbox.connect(
+                'refreshed', self._on_textbox_refreshed
+            )
+
     def _get_matches_for_unit(self, unit):
         return [match for match in self.matches if match.unit is unit]
 
@@ -274,51 +299,67 @@ class SearchMode(BaseMode):
         return (start, end)
 
     def _highlight_matches(self):
-        self._unhighlight_previous_matches()
-
         if not hasattr(self, 'filter') or not hasattr(self.filter, 're_search') or self.filter.re_search is None:
             return
 
-        unitview = self.controller.main_controller.unit_controller.view
-        self._prev_unitview = unitview
-        for textview in unitview.sources + unitview.targets:
-            buff = textview.get_buffer()
-            buffstr = buff.get_text(buff.get_start_iter(), buff.get_end_iter()).decode('utf-8')
-            unescaped = markup.unescape(buffstr)
+        for textbox in self.unitview.sources + self.unitview.targets:
+            self._highlight_textbox_matches(textbox)
 
-            # First make sure that the current buffer contains a highlighting tag.
-            # Because a gtk.TextTag can only be associated with one gtk.TagTable,
-            # we make copies (created by _make_highlight_tag()) to add to all
-            # TagTables. If the tag is already added to a given table, a
-            # ValueError is raised which we can safely ignore.
-            try:
-                buff.get_tag_table().add(self._make_highlight_tag())
-            except ValueError:
-                pass
+    def _get_matches_for_textbox(self, textbox):
+        if textbox.role == 'source':
+            textbox_n = self.unitview.sources.index(textbox)
+        elif textbox.role == 'target':
+            textbox_n = self.unitview.targets.index(textbox)
+        else:
+            raise ValueError('Could not find text box in sources or targets: %s' % (textbox))
+        return [
+            m for m in self.matches
+            if m.unit is self.unitview.unit and \
+                m.part == textbox.role and \
+                m.part_n == textbox_n
+        ]
 
-            select_iters = []
-            for match in [m for m in self.matches if m.unit is unitview.unit]:
-                if  (textview in unitview.sources and not match.part == 'source') or \
-                    (textview in unitview.targets and not match.part == 'target'):
-                    continue
-                start, end = self._escaped_indexes(unescaped, match.start, match.end)
-                start_iter, end_iter = buff.get_iter_at_offset(start), buff.get_iter_at_offset(end)
-                buff.apply_tag_by_name('highlight', start_iter, end_iter)
+    def _highlight_textbox_matches(self, textbox):
+        buff = textbox.buffer
+        buffstr = textbox.get_text()
+        unescaped = markup.unescape(buffstr)
 
-                if textview in unitview.targets and not select_iters and self.select_first_match:
-                    select_iters = [start_iter, end_iter]
+        # Make sure the 'search_highlight' tag in the textbox's tag table
+        # is "fresh".
+        try:
+            tagtable = buff.get_tag_table()
+            tag = tagtable.lookup('search_highlight')
+            if tag:
+                tagtable.remove(tag)
+            tagtable.add(self._make_highlight_tag())
+        except ValueError:
+            pass
 
-            if select_iters:
-                buff.select_range(select_iters[1], select_iters[0])
-                return
+        select_iters = []
+        for match in self._get_matches_for_textbox(textbox):
+            start, end = self._escaped_indexes(unescaped, match.start, match.end)
+            if hasattr(textbox.elem, 'gui_info'):
+                start = textbox.elem.gui_info.tree_to_gui_index(start)
+                end   = textbox.elem.gui_info.tree_to_gui_index(end)
+            start_iter, end_iter = buff.get_iter_at_offset(start), buff.get_iter_at_offset(end)
+            buff.apply_tag_by_name('search_highlight', start_iter, end_iter)
+
+            if textbox.role == 'target' and not select_iters and self.select_first_match:
+                select_iters = [start_iter, end_iter]
+
+        if select_iters:
+            buff.select_range(select_iters[1], select_iters[0])
 
     def _make_highlight_tag(self):
-        tag = gtk.TextTag(name='highlight')
+        tag = gtk.TextTag(name='search_highlight')
         tag.set_property('background', 'yellow')
         tag.set_property('foreground', 'black')
         return tag
 
     def _move_match(self, offset):
+        if self.controller.current_mode is not self:
+            return
+
         if getattr(self, 'matchcursor', None) is None:
             self.update_search()
             self._move_match(offset)
@@ -345,14 +386,6 @@ class SearchMode(BaseMode):
         self.controller.main_controller.undo_controller.record_stop()
         self.update_search()
 
-    def _unhighlight_previous_matches(self):
-        if not getattr(self, '_prev_unitview', ''):
-            return
-
-        for textview in self._prev_unitview.sources + self._prev_unitview.targets:
-            buff = textview.get_buffer()
-            buff.remove_all_tags(buff.get_start_iter(), buff.get_end_iter())
-
 
     # EVENT HANDLERS #
     def _on_entry_activate(self, entry):
@@ -414,6 +447,13 @@ class SearchMode(BaseMode):
         """This is called via the accelerator."""
         self.controller.select_mode(self)
 
+    def _on_textbox_refreshed(self, textbox, elem):
+        """Redoes highlighting after a C{StringElem} render destoyed it."""
+        if not textbox.props.visible or not unicode(elem):
+            return
+
+        self._highlight_textbox_matches(textbox)
+
     def _on_unit_modified(self, unit_controller, current_unit):
         unit_matches = self._get_matches_for_unit(current_unit)
         for match in unit_matches:
