@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright 2008 Zuza Software Foundation
+# Copyright 2008-2009 Zuza Software Foundation
 #
 # This file is part of Virtaal.
 #
@@ -127,7 +127,7 @@ class MainView(BaseView):
             try:
                 import igemacintegration
                 # Move the menu bar to the mac menu
-                menubar = self.gui.get_widget('menubar1')
+                menubar = self.gui.get_widget('menubar')
                 igemacintegration.ige_mac_menu_set_menu_bar(menubar)
                 menubar.hide()
                 igemacintegration.ige_mac_menu_connect_window_key_handler(self.main_window)
@@ -152,23 +152,24 @@ class MainView(BaseView):
 
         # Create our events dictionary and connect it
         dic = {
-                "on_mainwindow_destroy" : self._on_quit,
-                "on_mainwindow_delete" : self._on_quit,
-                "on_open_activate" : self._on_file_open,
-                "on_save_activate" : self._on_file_save,
-                "on_saveas_activate" : self._on_file_saveas,
-                "on_update_activate" : self._on_file_update,
-                "on_revert_activate" : self._on_file_revert,
-                "on_quit" : self._on_quit,
-                "on_menuitem_fullscreen_toggled" : self._on_fullscreen,
-                "on_about_activate" : self._on_help_about,
-                "on_tutorial_activate" : self._on_tutorial,
-                "on_localization_guide_activate" : self._on_localization_guide,
-                "on_menuitem_documentation_activate" : self._on_documentation,
-                "on_menuitem_report_bug_activate" : self._on_report_bug,
-                }
+            "on_mainwindow_destroy" : self._on_quit,
+            "on_mainwindow_delete" : self._on_quit,
+            "on_open_activate" : self._on_file_open,
+            "on_save_activate" : self._on_file_save,
+            "on_saveas_activate" : self._on_file_saveas,
+            "on_update_activate" : self._on_file_update,
+            "on_revert_activate" : self._on_file_revert,
+            "on_quit" : self._on_quit,
+            "on_menuitem_fullscreen_toggled" : self._on_fullscreen,
+            "on_about_activate" : self._on_help_about,
+            "on_tutorial_activate" : self._on_tutorial,
+            "on_localization_guide_activate" : self._on_localization_guide,
+            "on_menuitem_documentation_activate" : self._on_documentation,
+            "on_menuitem_report_bug_activate" : self._on_report_bug,
+        }
         self.gui.signal_autoconnect(dic)
 
+        self.menubar = self.gui.get_widget('menubar')
         self.status_bar = self.gui.get_widget("status_bar")
         self.status_bar.set_sensitive(False)
         self.statusbar_context_id = self.status_bar.get_context_id("statusbar")
@@ -188,6 +189,7 @@ class MainView(BaseView):
         self.controller.connect('controller-registered', self._on_controller_registered)
         self._create_dialogs()
         self._setup_key_bindings()
+        self._track_window_state()
 
     def _create_dialogs(self):
         self.input_dialog = EntryDialog(self.main_window)
@@ -212,7 +214,7 @@ class MainView(BaseView):
         )
 
         self.open_chooser = gtk.FileChooserDialog(
-            _('Choose a translation file'),
+            _('Choose a Translation File'),
             self.main_window,
             gtk.FILE_CHOOSER_ACTION_OPEN,
             (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK)
@@ -266,20 +268,27 @@ class MainView(BaseView):
             gtk.BUTTONS_NONE,
             _("The current file has been modified.\nDo you want to save your changes?")
         )
-        self.confirm_dialog.add_buttons(gtk.STOCK_SAVE, RESPONSE_SAVE)
-        self.confirm_dialog.add_buttons(_("_Discard"), RESPONSE_DISCARD)
-        self.confirm_dialog.add_buttons(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
+        self.confirm_dialog.__save_button = self.confirm_dialog.add_button(gtk.STOCK_SAVE, RESPONSE_SAVE)
+        self.confirm_dialog.add_button(_("_Discard"), RESPONSE_DISCARD)
+        self.confirm_dialog.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
         self.confirm_dialog.set_default_response(RESPONSE_SAVE)
 
     def _setup_key_bindings(self):
         self.accel_group = gtk.AccelGroup()
         self.main_window.add_accel_group(self.accel_group)
 
+    def _track_window_state(self):
+        self._window_is_maximized = False
+
+        def on_state_event(widget, event):
+            self._window_is_maximized = bool(event.new_window_state & gtk.gdk.WINDOW_STATE_MAXIMIZED)
+        self.main_window.connect('window-state-event', on_state_event)
+
 
     # ACCESSORS #
     def add_accel_group(self, accel_group):
         """Add the given accelerator group to the main window.
-            @type accel_group: gtk.AccelGroup()"""
+            @type accel_group: gtk.AccelGroup"""
         self.main_window.add_accel_group(accel_group)
 
     def set_saveable(self, value):
@@ -293,7 +302,11 @@ class MainView(BaseView):
             if value:
                 modified = "*"
             self.main_window.set_title(
-                    (_('Virtaal - %(current_file)s %(modified_marker)s') %
+                    #l10n: This is the title of the main window of Virtaal
+                    #%(modified_marker)s is a star that is displayed if the file is modified, and should be at the start of the window title
+                    #%(current_file)s is the file name of the current file
+                    #most languages will not need to change this
+                    (_('%(modified_marker)s%(current_file)s - Virtaal') %
                         {
                             "current_file": os.path.basename(filename),
                             "modified_marker": modified
@@ -335,14 +348,87 @@ class MainView(BaseView):
             # Note that if nplurals == 1, the default equation "0" is correct
             return nplurals, lang.pluralequation
 
+    def append_menu(self, name):
+        """Add a menu with the given name to the menu bar."""
+        menu = gtk.Menu()
+        menuitem = gtk.MenuItem(name)
+        menuitem.set_submenu(menu)
+        self.menubar.append(menuitem)
+        self.menubar.show_all()
+        return menuitem
+
+    def append_menu_item(self, name, menu, after=None):
+        """Add a new menu item with the given name to the menu with the given
+            name (C{menu})."""
+        if isinstance(after, (str, unicode)):
+            after = self.find_menu(after)
+
+        parent_item = None
+        if isinstance(menu, gtk.MenuItem):
+            parent_item = menu
+        else:
+            parent_item = self.find_menu(menu)
+        if parent_item is None:
+            return None
+
+        parent_menu = parent_item.get_submenu()
+        menuitem = gtk.MenuItem(name)
+        if after is None:
+            parent_menu.add(menuitem)
+        else:
+            after_index = parent_menu.get_children().index(after) + 1
+            parent_menu.insert(menuitem, after_index)
+        self.menubar.show_all()
+        return menuitem
+
+    def find_menu(self, label):
+        """Find the menu with the given label on the menu bar."""
+        for menuitem in self.menubar.get_children():
+            if menuitem.get_child() and menuitem.get_child().get_text() == label:
+                return menuitem
+
+        if '_' in label:
+            return self.find_menu(label.replace('_', ''))
+
+        return None
+
+    def find_menu_item(self, label, menu=None):
+        """Find the menu item with the given label and in the menu with the
+            given name (if it exists).
+
+            @param label: The label of the menu item to find.
+            @param menu: The (optional) (name of the) menu to search in."""
+        if not isinstance(menu, gtk.MenuItem):
+            menu = self.find_menu(label)
+        if menu is not None:
+            menus = [menu]
+        else:
+            menus = [mi for mi in self.menubar.get_children()]
+
+        for menuitem in menus:
+            for item in menuitem.get_submenu().get_children():
+                if item.get_child() and item.get_child().get_text() == label:
+                    return item, menuitem
+
+        if '_' in label:
+            return self.find_menu_item(label.replace('_', ''), menu)
+
+        return None, None
+
     def quit(self):
-        width, height = self.main_window.get_size()
-        pan_app.settings.general['windowwidth'] = width
-        pan_app.settings.general['windowheight'] = height
+        if self._window_is_maximized:
+            pan_app.settings.general['maximized'] = 1
+        else:
+            width, height = self.main_window.get_size()
+            pan_app.settings.general['windowwidth'] = width
+            pan_app.settings.general['windowheight'] = height
+            pan_app.settings.general['maximized'] = ''
         pan_app.settings.write()
         gtk.main_quit()
 
     def show(self):
+        if pan_app.settings.general['maximized']:
+            self.main_window.maximize()
         self.main_window.show()
         gtk.main()
 
@@ -440,6 +526,7 @@ class MainView(BaseView):
         self.confirm_dialog.set_transient_for(self._top_window)
         old_top = self._top_window
         self._top_window = self.confirm_dialog
+        self.confirm_dialog.__save_button.grab_focus()
         response = self.confirm_dialog.run()
         self.confirm_dialog.hide()
         self._top_window = old_top
@@ -543,6 +630,8 @@ class MainView(BaseView):
         self.status_bar.set_sensitive(True)
         if getattr(self, '_uri', None):
             recent.rm.add_item(self._uri)
+        else:
+            recent.rm.add_item('file://' + os.path.abspath(store_controller.store.filename))
 
     def _on_window_state_event(self, widget, event):
         mnu_fullscreen = self.gui.get_widget('menuitem_fullscreen')
