@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright 2008 Zuza Software Foundation
+# Copyright 2008-2009 Zuza Software Foundation
 #
 # This file is part of Virtaal.
 #
@@ -21,9 +21,8 @@
 import gtk
 import logging
 import re
-from gobject import idle_add, GObject, SIGNAL_RUN_FIRST, TYPE_INT, TYPE_NONE, TYPE_PYOBJECT, TYPE_STRING
+from gobject import idle_add, GObject, SIGNAL_RUN_FIRST, TYPE_PYOBJECT
 from translate.lang import factory
-from translate.lang import data
 try:
     import gtkspell
 except ImportError, e:
@@ -35,6 +34,7 @@ import markup
 import rendering
 from baseview import BaseView
 from widgets.label_expander import LabelExpander
+from widgets.textbox import TextBox
 
 
 class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
@@ -42,12 +42,12 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
 
     __gtype_name__ = "UnitView"
     __gsignals__ = {
-        'delete-text':    (SIGNAL_RUN_FIRST, TYPE_NONE, (TYPE_STRING, TYPE_INT, TYPE_INT, TYPE_INT, TYPE_INT)),
-        'insert-text':    (SIGNAL_RUN_FIRST, TYPE_NONE, (TYPE_STRING, TYPE_STRING, TYPE_INT, TYPE_INT)),
-        'paste-start':    (SIGNAL_RUN_FIRST, TYPE_NONE, (TYPE_STRING, TYPE_PYOBJECT, TYPE_INT)),
-        'modified':       (SIGNAL_RUN_FIRST, TYPE_NONE, ()),
-        'unit-done':      (SIGNAL_RUN_FIRST, TYPE_NONE, (TYPE_PYOBJECT,)),
-        'target-focused': (SIGNAL_RUN_FIRST, TYPE_NONE, (TYPE_INT,)),
+        'delete-text':    (SIGNAL_RUN_FIRST, None, (TYPE_PYOBJECT, TYPE_PYOBJECT, int, int, TYPE_PYOBJECT, int)),
+        'insert-text':    (SIGNAL_RUN_FIRST, None, (TYPE_PYOBJECT, int, TYPE_PYOBJECT, int)),
+        'paste-start':    (SIGNAL_RUN_FIRST, None, (TYPE_PYOBJECT, TYPE_PYOBJECT, int)),
+        'modified':       (SIGNAL_RUN_FIRST, None, ()),
+        'unit-done':      (SIGNAL_RUN_FIRST, None, (TYPE_PYOBJECT,)),
+        'target-focused': (SIGNAL_RUN_FIRST, None, (int,)),
     }
 
     first_word_re = re.compile("(?m)(?u)^(<[^>]+>|\\\\[nt]|[\W$^\n])*(\\b|\\Z)")
@@ -86,18 +86,75 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         self.load_unit(unit)
 
     def _setup_menus(self):
+        def get_focused(widgets):
+            for textview in widgets:
+                if textview.is_focus():
+                    return textview
+            return None
+
         clipboard = gtk.Clipboard(selection=gtk.gdk.SELECTION_CLIPBOARD)
         def on_cut(menuitem):
-            self.targets[self.focused_target_n].get_buffer().cut_clipboard(clipboard, True)
+            focused = get_focused(self.targets)
+            if focused is not None:
+                focused.get_buffer().cut_clipboard(clipboard, True)
         def on_copy(menuitem):
-            self.targets[self.focused_target_n].get_buffer().copy_clipboard(clipboard)
+            focused = get_focused(self.targets + self.sources)
+            if focused is not None:
+                focused.get_buffer().copy_clipboard(clipboard)
         def on_paste(menuitem):
-            self.targets[self.focused_target_n].get_buffer().paste_clipboard(clipboard, None, True)
+            focused = get_focused(self.targets)
+            if focused is not None:
+                focused.get_buffer().paste_clipboard(clipboard, None, True)
 
         maingui = self.controller.main_controller.view.gui
-        maingui.get_widget('mnu_cut').connect('activate', on_cut)
-        maingui.get_widget('mnu_copy').connect('activate', on_copy)
-        maingui.get_widget('mnu_paste').connect('activate', on_paste)
+        self.mnu_cut = maingui.get_widget('mnu_cut')
+        self.mnu_copy = maingui.get_widget('mnu_copy')
+        self.mnu_paste = maingui.get_widget('mnu_paste')
+
+        self.mnu_cut.connect('activate', on_cut)
+        self.mnu_copy.connect('activate', on_copy)
+        self.mnu_paste.connect('activate', on_paste)
+
+        # And now for the "Transfer from source" and placeable selection menu items
+        mnu_next = maingui.get_widget('mnu_placnext')
+        mnu_prev = maingui.get_widget('mnu_placprev')
+        mnu_transfer = maingui.get_widget('mnu_transfer')
+        self.mnu_next = mnu_next
+        self.mnu_prev = mnu_prev
+        self.mnu_transfer = mnu_transfer
+        menu_edit = maingui.get_widget('menu_edit')
+
+        def on_next(*args):
+            self.targets[self.focused_target_n].move_elem_selection(1)
+        def on_prev(*args):
+            self.targets[self.focused_target_n].move_elem_selection(-1)
+        def on_transfer(*args):
+            ev = gtk.gdk.Event(gtk.gdk.KEY_PRESS)
+            ev.state = gtk.gdk.MOD1_MASK
+            ev.keyval = gtk.keysyms.Down
+            ev.window = self.targets[self.focused_target_n].get_window(gtk.TEXT_WINDOW_WIDGET)
+            ev.put()
+        mnu_next.connect('activate', on_next)
+        mnu_prev.connect('activate', on_prev)
+        mnu_transfer.connect('activate', on_transfer)
+
+        gtk.accel_map_add_entry("<Virtaal>/Edit/Next Placeable", gtk.keysyms.Right, gtk.gdk.MOD1_MASK)
+        gtk.accel_map_add_entry("<Virtaal>/Edit/Prev Placeable", gtk.keysyms.Left, gtk.gdk.MOD1_MASK)
+        gtk.accel_map_add_entry("<Virtaal>/Edit/Transfer", gtk.keysyms.Down, gtk.gdk.MOD1_MASK)
+
+        accel_group = menu_edit.get_accel_group()
+        if not accel_group:
+            accel_group = gtk.AccelGroup()
+
+        self.controller.main_controller.view.add_accel_group(accel_group)
+        menu_edit.set_accel_group(accel_group)
+        mnu_next.set_accel_path("<Virtaal>/Edit/Next Placeable")
+        mnu_prev.set_accel_path("<Virtaal>/Edit/Prev Placeable")
+        mnu_transfer.set_accel_path("<Virtaal>/Edit/Transfer")
+
+        # Disable the menu items to start with, because we can't assume that a
+        # store is loaded. See _set_menu_items_sensitive() for more activation.
+        self._set_menu_items_sensitive(False)
 
 
     # ACCESSORS #
@@ -111,43 +168,54 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
     focused_target_n = property(_get_focused_target_n, _set_focused_target_n)
 
     def get_target_n(self, n):
-        buff = self.targets[n].get_buffer()
-        return data.forceunicode(markup.unescape(buff.get_text(buff.get_start_iter(), buff.get_end_iter())))
+        return self.targets[n].get_text()
 
-    def set_target_n(self, n, newtext, cursor_pos=-1, escape=True):
+    def set_target_n(self, n, newtext, cursor_pos=-1):
         # TODO: Save cursor position and set after assignment
-        buff = self.targets[n].get_buffer()
-        if escape:
-            newtext = markup.escape(newtext)
-        buff.set_text(newtext)
+        self.targets[n].set_text(newtext)
         if cursor_pos > -1:
-            buff.place_cursor(buff.get_iter_at_offset(cursor_pos))
+            self.targets[n].buffer.place_cursor(self.targets[n].buffer.get_iter_at_offset(cursor_pos))
 
     sources = property(lambda self: self._widgets['sources'])
     targets = property(lambda self: self._widgets['targets'])
 
 
     # METHODS #
-    def copy_original(self, text_view):
-        buf = text_view.get_buffer()
-        position = buf.props.cursor_position
-        old_text = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
+    def copy_original(self, textbox):
+        if textbox.selector_textbox is not textbox and \
+            textbox.selector_textbox.selected_elem is not None:
+            textbox.insert_translation(textbox.selector_textbox.selected_elem)
+            textbox.selector_textbox.move_elem_selection(1)
+            return
+
+        undocontroller = self.controller.main_controller.undo_controller
         lang = factory.getlanguage(self.controller.main_controller.lang_controller.target_lang.code)
-        new_source = lang.punctranslate(self.unit.source)
-        # if punctranslate actually changed something, let's insert that as an
-        # undo step
-        if new_source != self.unit.source:
-            self.controller.main_controller.undo_controller.record_start()
-            buf.set_text(markup.escape(self.unit.source))
-            self.controller.main_controller.undo_controller.record_stop()
 
-        self.controller.main_controller.undo_controller.record_start()
-        buf.set_text(markup.escape(new_source))
-        self.controller.main_controller.undo_controller.record_stop()
+        tgt = self.unit.rich_source[0].copy()
+        placeables_controller = self.controller.main_controller.placeables_controller
+        parsers = placeables_controller.get_parsers_for_textbox(textbox)
+        placeables_controller.apply_parsers(tgt, parsers)
+        if textbox.role == 'target':
+            for plac in placeables_controller.non_target_placeables:
+                tgt.remove_type(plac)
+        tgt.prune()
 
-        # The following 2 lines were copied from focus_text_view() below
-        translation_start = self.first_word_re.match(markup.escape(new_source)).span()[1]
-        buf.place_cursor(buf.get_iter_at_offset(translation_start))
+        punctgt = tgt.copy()
+        punctgt.map(
+            lambda e: e.apply_to_strings(lang.punctranslate),
+            lambda e: e.isleaf() and e.istranslatable
+        )
+
+        if punctgt != tgt:
+            undocontroller.push_current_text(textbox)
+            textbox.set_text(tgt)
+            tgt = punctgt
+
+        undocontroller.push_current_text(textbox)
+        textbox.set_text(tgt)
+
+        textbox.refresh_cursor_pos = self._get_editing_start_pos(textbox.elem)
+        textbox.refresh()
 
         return False
 
@@ -156,20 +224,16 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         self.focus_text_view(self.targets[0])
 
     def do_editing_done(self, *_args):
-        #logging.debug('emit("unit-done", self.unit=%s)' % (self.unit))
-        #self.emit('unit-done', self.unit)
         pass
 
-    def focus_text_view(self, text_view):
-        text_view.grab_focus()
+    def focus_text_view(self, textbox):
+        textbox.grab_focus()
 
-        buf = text_view.get_buffer()
-        text = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
+        text = textbox.get_text()
+        translation_start = self._get_editing_start_pos(textbox.elem)
+        textbox.buffer.place_cursor(textbox.buffer.get_iter_at_offset(translation_start))
 
-        translation_start = self.first_word_re.match(text).span()[1]
-        buf.place_cursor(buf.get_iter_at_offset(translation_start))
-
-        self._focused_target_n = self.targets.index(text_view)
+        self._focused_target_n = self.targets.index(textbox)
         #logging.debug('emit("target-focused", focused_target_n=%d)' % (self._focused_target_n))
         self.emit('target-focused', self._focused_target_n)
 
@@ -177,6 +241,13 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         """Load a GUI (C{gtk.CellEditable}) for the given unit."""
         if unit is self.unit and unit is not None:
             return
+
+        if self.unit is not None:
+            #logging.debug('emit("unit-done", self.unit=%s)' % (self.unit))
+            self.emit('unit-done', self.unit)
+            for src in self.sources:
+                src.select_elem(elem=None)
+
         self.unit = unit
         self.disable_signals(['modified', 'insert-text', 'delete-text'])
         self._update_editor_gui()
@@ -211,6 +282,20 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
             textview.modify_font(rendering.get_target_font_description())
             textview.get_pango_context().set_font_description(rendering.get_target_font_description())
 
+    def _get_editing_start_pos(self, elem):
+        if not elem:
+            return 0
+        translation_start = self.first_word_re.match(unicode(elem)).span()[1]
+        start_elem = elem.elem_at_offset(translation_start)
+        if not start_elem.iseditable:
+            flattened = elem.flatten()
+            start_index = flattened.index(start_elem)
+            if start_index == len(flattened)-1:
+                return len(elem)
+            next_elem = flattened[start_index+1]
+            return elem.elem_offset(next_elem)
+        return translation_start
+
     def _get_widgets(self):
         """Get the widgets we would like to use from the loaded Glade XML object."""
         if not getattr(self, '_widgets', None):
@@ -223,20 +308,26 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
 
         self._widgets['vbox_targets'].connect('key-press-event', self._on_key_press_event)
 
+    def _set_menu_items_sensitive(self, sensitive=True):
+        for widget in (self.mnu_next, self.mnu_prev, self.mnu_transfer):
+            widget.set_sensitive(sensitive)
+
     def _update_editor_gui(self):
         """Build the default editor with the following components:
             - A C{gtk.TextView} for each source
             - A C{gtk.TextView} for each target
             - A C{gtk.ToggleButton} for the fuzzy option
-            - A C{widgets.LabelExpander} for programmer notes
-            - A C{widgets.LabelExpander} for translator notes
-            - A C{widgets.LabelExpander} for context info"""
+            - A C{gtk.Label} for programmer notes
+            - A C{gtk.Label} for translator notes
+            - A C{gtk.Label} for context info"""
         self._layout_update_notes('programmer')
         self._layout_update_sources()
         self._layout_update_context_info()
         self._layout_update_targets()
         self._layout_update_notes('translator')
         self._layout_update_fuzzy()
+        if self.unit:
+            self._set_menu_items_sensitive(True)
 
     def _update_textview_language(self, text_view, language):
         language = str(language)
@@ -309,80 +400,77 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
     # GUI BUILDING CODE #
     def _create_sources(self):
         for i in range(len(self.sources), self.MAX_SOURCES):
-            source = self._create_textbox('', editable=False, scroll_policy=gtk.POLICY_NEVER)
-            textview = source.get_child()
+            source = self._create_textbox(u'', editable=False, role='source')
+            textbox = source.get_child()
+            textbox.modify_font(rendering.get_source_font_description())
             self._widgets['vbox_sources'].pack_start(source)
-            self.sources.append(textview)
+            self.sources.append(textbox)
 
             # The following fixes a very weird crash (bug #810)
-            def ignore_tab(text_view, event):
+            def ignore_tab(txtbx, event, eventname):
                 if event.keyval in (gtk.keysyms.Tab, gtk.keysyms.ISO_Left_Tab):
                     self.focused_target_n = 0
                     return True
-            textview.connect('key-press-event', ignore_tab)
+            textbox.connect('key-pressed', ignore_tab)
 
     def _create_targets(self):
-        def on_text_view_n_press_event(text_view, event):
+        def on_textbox_n_press_event(textbox, event, eventname):
             """Handle special keypresses in the textarea."""
-            # Alt-Down
-            if event.keyval == gtk.keysyms.Down and event.state & gtk.gdk.MOD1_MASK:
-                idle_add(self.copy_original, text_view)
-                return True
 
-            # Automatically move to the next line if \n is entered
-            if event.keyval == gtk.keysyms.n:
-                buf = text_view.get_buffer()
-                curpos = buf.props.cursor_position
-                lastchar = buf.get_text(
-                    buf.get_iter_at_offset(curpos-1),
-                    buf.get_iter_at_offset(curpos)
-                )
-                if lastchar == "\\":
-                    buf.insert_at_cursor('n\n')
-                    text_view.scroll_mark_onscreen(buf.get_insert())
-                    return True
-
-            return False
-
-        def target_key_press_event(text_view, event, next_text_view):
-            if event.keyval == gtk.keysyms.Return or event.keyval == gtk.keysyms.KP_Enter:
-                if next_text_view is not None and next_text_view.props.visible:
-                    self.focus_text_view(next_text_view)
+        def target_key_press_event(textbox, event, eventname, next_textbox):
+            if eventname == 'enter':
+                if next_textbox is not None and next_textbox.props.visible:
+                    self.focus_text_view(next_textbox)
                 else:
-                    # text_view is the last text view in this unit, so we need to move on
+                    # textbox is the last text view in this unit, so we need to move on
                     # to the next one.
-                    text_view.parent.parent.emit('key-press-event', event)
+                    textbox.parent.parent.emit('key-press-event', event)
                 return True
+
+            # Alt-Down
+            elif eventname == 'alt-down':
+                idle_add(self.copy_original, textbox)
+                return True
+
+            # Shift-Tab
+            elif eventname == 'shift-tab':
+                if self.focused_target_n > 0:
+                    self.focused_target_n -= 1
+                return True
+
             return False
 
         for i in range(len(self.targets), self.MAX_TARGETS):
-            target = self._create_textbox('', editable=True, scroll_policy=gtk.POLICY_AUTOMATIC)
-            textview = target.get_child()
-            buff = textview.get_buffer()
-            textview.connect('key-press-event', on_text_view_n_press_event)
-            textview.connect('paste-clipboard', self._on_textview_paste_clipboard, i)
-            buff.connect('changed', self._on_target_changed, i)
-            buff.connect('insert-text', self._on_target_insert_text, i)
-            buff.connect('delete-range', self._on_target_delete_range, i)
+            target = self._create_textbox(u'', editable=True, role='target', scroll_policy=gtk.POLICY_AUTOMATIC)
+            textbox = target.get_child()
+            textbox.modify_font(rendering.get_target_font_description())
+            textbox.selector_textbox = self.sources[0]
+            textbox.connect('paste-clipboard', self._on_textbox_paste_clipboard, i)
+            textbox.connect('text-inserted', self._on_target_insert_text, i)
+            textbox.connect('text-deleted', self._on_target_delete_range, i)
+            textbox.buffer.connect('changed', self._on_target_changed, i)
 
             self._widgets['vbox_targets'].pack_start(target)
-            self.targets.append(textview)
+            self.targets.append(textbox)
 
         for target, next_target in zip(self.targets, self.targets[1:] + [None]):
-            target.connect('key-press-event', target_key_press_event, next_target)
+            target.connect('key-pressed', target_key_press_event, next_target)
 
-    def _create_textbox(self, text='', editable=True, scroll_policy=gtk.POLICY_AUTOMATIC):
-        textview = gtk.TextView()
-        textview.set_editable(editable)
-        textview.set_wrap_mode(gtk.WRAP_WORD_CHAR)
-        textview.set_border_window_size(gtk.TEXT_WINDOW_TOP, 1)
-        textview.set_left_margin(2)
-        textview.set_right_margin(2)
-        textview.get_buffer().set_text(text or '')
+    def _create_textbox(self, text=u'', editable=True, role=None, scroll_policy=gtk.POLICY_AUTOMATIC):
+        textbox = TextBox(self.controller.main_controller, role=role)
+        textbox.set_editable(editable)
+        textbox.set_wrap_mode(gtk.WRAP_WORD_CHAR)
+        textbox.set_border_window_size(gtk.TEXT_WINDOW_TOP, 1)
+        textbox.set_left_margin(2)
+        textbox.set_right_margin(2)
+        textbox.set_text(text or u'')
+        textbox.connect('focus-in-event', self._on_textbox_focused)
+        textbox.connect('focus-out-event', self._on_textbox_unfocused)
 
         scrollwnd = gtk.ScrolledWindow()
         scrollwnd.set_policy(gtk.POLICY_NEVER, scroll_policy)
-        scrollwnd.add(textview)
+        scrollwnd.set_shadow_type(gtk.SHADOW_IN)
+        scrollwnd.add(textbox)
 
         return scrollwnd
 
@@ -390,6 +478,8 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         if origin not in self._widgets['notes']:
             label = gtk.Label()
             label.set_line_wrap(True)
+            label.set_justify(gtk.JUSTIFY_FILL)
+            label.set_property('selectable', True)
 
             self._widgets['vbox_middle'].pack_start(label)
             if origin == 'programmer':
@@ -407,6 +497,15 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         if origin == "programmer" and len(note_text) < 15 and self.unit is not None and self.unit.getlocations():
             note_text += u"  " + u" ".join(self.unit.getlocations()[:3])
 
+        # FIXME: This is a temporary quick fix (to bug 1145) to ensure that
+        # excessive translator comments don't cover the whole display.
+        # The labels used for displaying these comments (programmer- as well as
+        # translator comments) should be displayed in a scrollable widget with
+        # proper size limitations.
+        TEXT_LIMIT = 200
+        if origin == "translator" and len(note_text) > TEXT_LIMIT:
+            note_text = note_text[:TEXT_LIMIT] + '...'
+
         self._widgets['notes'][origin].set_text(note_text)
 
         if note_text:
@@ -426,9 +525,9 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         if self.unit is None:
             if num_source_widgets >= 1:
                 # The above condition should *never* be False
-                txtview = self.sources[0]
-                txtview.get_buffer().set_text('')
-                txtview.parent.show()
+                textbox = self.sources[0]
+                textbox.set_text(u'')
+                textbox.parent.show()
             for i in range(1, num_source_widgets):
                 self.sources[i].parent.hide_all()
             return
@@ -439,16 +538,9 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
 
         for i in range(self.MAX_SOURCES):
             if i < num_unit_sources:
-                sourcestr = ''
-                if self.unit.hasplural():
-                    sourcestr = self.unit.source.strings[i]
-                elif i == 0:
-                    sourcestr = self.unit.source
-                else:
-                    raise IndexError()
-
+                sourcestr = self.unit.rich_source[i]
                 self.sources[i].modify_font(rendering.get_source_font_description())
-                self.sources[i].get_buffer().set_text(markup.escape(sourcestr))
+                self.sources[i].set_text(sourcestr)
                 self.sources[i].parent.show_all()
                 #logging.debug('Showing source #%d: %s' % (i, self.sources[i]))
             else:
@@ -461,21 +553,19 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
                 self._widgets['context_info'].hide()
             return
 
-        if self._widgets['context_info']:
-            self._widgets['context_info'].show()
-            self._widgets['context_info'].get_buffer().set_text(self.unit.getcontext() or u"")
-        else:
-            textbox = self._create_textbox(
-                self.unit.getcontext(),
-                editable=False,
-                scroll_policy=gtk.POLICY_NEVER
-            )
-            textview = textbox.get_child()
-            labelexpander = LabelExpander(textbox, lambda *args: self.unit.getcontext())
+        if not self._widgets['context_info']:
+            label = gtk.Label()
+            label.set_line_wrap(True)
+            label.set_justify(gtk.JUSTIFY_FILL)
+            self._widgets['vbox_middle'].pack_start(label)
+            self._widgets['vbox_middle'].reorder_child(label, 2)
+            self._widgets['context_info'] = label
 
-            self._widgets['vbox_middle'].pack_start(labelexpander)
-            self._widgets['vbox_middle'].reorder_child(labelexpander, 2)
-            self._widgets['context_info'] = textview
+        if self.unit.getcontext():
+            self._widgets['context_info'].show()
+            self._widgets['context_info'].set_text(self.unit.getcontext() or u"")
+        else:
+            self._widgets['context_info'].hide()
 
     def _layout_update_targets(self):
         num_target_widgets = len(self.targets)
@@ -489,9 +579,9 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
         if self.unit is None:
             if num_target_widgets >= 1:
                 # The above condition should *never* be False
-                txtview = self.targets[0]
-                txtview.get_buffer().set_text('')
-                txtview.parent.show_all()
+                textbox = self.targets[0]
+                textbox.set_text(u'')
+                textbox.parent.show_all()
             for i in range(1, num_target_widgets):
                 self.targets[i].parent.hide_all()
             return
@@ -502,18 +592,16 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
             num_unit_targets = len(self.unit.target.strings)
             nplurals = self.controller.main_controller.lang_controller.target_lang.nplurals
 
+        rich_target = self.unit.rich_target
+        rich_target_len = len(rich_target)
         for i in range(self.MAX_TARGETS):
             if i < nplurals:
                 # plural forms already in file
-                if self.unit.hasplural() and i < num_unit_targets:
-                    targetstr = self.unit.target.strings[i]
-                elif i == 0:
-                    targetstr = self.unit.target
-                else:
-                    targetstr = ''
-
+                targetstr = u''
+                if i < rich_target_len and rich_target[i] is not None:
+                    targetstr = rich_target[i]
                 self.targets[i].modify_font(rendering.get_target_font_description())
-                self.targets[i].get_buffer().set_text(markup.escape(targetstr))
+                self.targets[i].set_text(targetstr)
                 self.targets[i].parent.show_all()
                 #logging.debug('Showing target #%d: %s' % (i, self.targets[i]))
             else:
@@ -524,6 +612,7 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
     def _layout_update_fuzzy(self):
         if not self._widgets['fuzzy']:
             fuzzy = gtk.CheckButton(label=_('F_uzzy'))
+            fuzzy.set_property("xalign", 0.0)
             # FIXME: not allowing focus will probably raise various issues related to keyboard accesss.
             fuzzy.set_property("can-focus", False)
             fuzzy.connect('toggled', self._on_fuzzy_toggled)
@@ -545,50 +634,50 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
     def _on_key_press_event(self, _widget, event, *_args):
         if event.keyval == gtk.keysyms.Return or event.keyval == gtk.keysyms.KP_Enter:
             self.must_advance = True
-            self.emit('unit-done', self.unit)
+            # Clear selected elements
             self.editing_done()
             return True
         return False
 
     def _on_target_changed(self, buffer, index):
-        newtext = self.get_target_n(index)
-        if self.unit.hasplural():
-            nplurals = self.controller.main_controller.lang_controller.target_lang.nplurals
-            # FIXME: The following two lines are necessary because self.unit.target always
-            # returns a new multistring, so you can't assign to an index directly.
-            target = self.unit.target.strings
-            if len(target) < nplurals:
+        tgt = self.targets[index]
+        nplurals = self.controller.main_controller.lang_controller.target_lang.nplurals
+        if tgt.elem is not None:
+            rich_target = self.unit.rich_target
+            if len(rich_target) < nplurals:
                 # pad the target with empty strings
-                target += (nplurals - len(target)) * [u""]
-            target[index] = newtext
-            self.unit.target = target
-        elif index == 0:
-            self.unit.target = newtext
+                rich_target += (nplurals - len(rich_target)) * [u""]
+            rich_target[index] = tgt.elem
+            self.unit.rich_target = rich_target
         else:
-            raise IndexError()
+            newtext = self.get_target_n(index)
+            if self.unit.hasplural():
+                # FIXME: The following two lines are necessary because self.unit.target always
+                # returns a new multistring, so you can't assign to an index directly.
+                target = self.unit.target.strings
+                if len(target) < nplurals:
+                    # pad the target with empty strings
+                    target += (nplurals - len(target)) * [u""]
+                target[index] = newtext
+                self.unit.target = target
+            elif index == 0:
+                self.unit.target = newtext
+            else:
+                raise IndexError()
 
         self.modified()
 
-    def _on_target_insert_text(self, buff, iter, ins_text, length, target_num):
-        old_text = buff.get_text(buff.get_start_iter(), buff.get_end_iter())
-        offset = len(buff.get_text(buff.get_start_iter(), iter)) # FIXME: Isn't there a better way to do this?
+    def _on_target_insert_text(self, textbox, ins_text, offset, elem, target_num):
+        #logging.debug('emit("insert-text", ins_text="%s", offset=%d, elem=%s, target_num=%d)' % (ins_text, offset, repr(elem), target_num))
+        self.emit('insert-text', ins_text, offset, elem, target_num)
 
-        #logging.debug('emit("insert-text", old_text="%s", ins_text="%s", offset=%d, target_num=%d)' % (old_text, ins_text, offset, target_num))
-        self.emit('insert-text', old_text, ins_text, offset, target_num)
+    def _on_target_delete_range(self, textbox, deleted, parent, offset, cursor_pos, elem, target_num):
+        #logging.debug('emit("delete-text", start_offset=%d, end_offset=%d, cursor_pos=%d, elem=%s, target_num=%d)' % (old_text, start_offset, end_offset, cursor_pos, repr(elem), target_num))
+        self.emit('delete-text', deleted, parent, offset, cursor_pos, elem, target_num)
 
-    def _on_target_delete_range(self, buff, start_iter, end_iter, target_num):
-        cursor_iter = buff.get_iter_at_mark(buff.get_insert())
-        cursor_pos = len(buff.get_text(buff.get_start_iter(), cursor_iter))
-        old_text = buff.get_text(buff.get_start_iter(), buff.get_end_iter())
-        start_offset = len(buff.get_text(buff.get_start_iter(), start_iter)) # FIXME: Isn't there a better way to do this?
-        end_offset = len(buff.get_text(buff.get_start_iter(), end_iter)) # FIXME: Isn't there a better way to do this?
-
-        #logging.debug('emit("delete-text", old_text="%s", start_offset=%d, end_offset=%d, cursor_pos=%d, target_num=%d)' % (old_text, start_offset, end_offset, cursor_pos, target_num))
-        self.emit('delete-text', old_text, start_offset, end_offset, cursor_pos, target_num)
-
-    def _on_textview_paste_clipboard(self, textview, target_num):
-        buff = textview.get_buffer()
-        old_text = buff.get_text(buff.get_start_iter(), buff.get_end_iter())
+    def _on_textbox_paste_clipboard(self, textbox, target_num):
+        buff = textbox.buffer
+        old_text = textbox.get_text()
         ins_iter  = buff.get_iter_at_mark(buff.get_insert())
         selb_iter = buff.get_iter_at_mark(buff.get_selection_bound())
 
@@ -599,3 +688,11 @@ class UnitView(gtk.EventBox, GObjectWrapper, gtk.CellEditable, BaseView):
 
         #logging.debug('emit("paste-start", old_text="%s", offsets=%d, target_num=%d)' % (old_text, offsets, target_num))
         self.emit('paste-start', old_text, offsets, target_num)
+
+    def _on_textbox_focused(self, textbox, event):
+        for mnu in (self.mnu_cut, self.mnu_copy, self.mnu_paste):
+            mnu.set_sensitive(True)
+
+    def _on_textbox_unfocused(self, textbox, event):
+        for mnu in (self.mnu_cut, self.mnu_copy, self.mnu_paste):
+            mnu.set_sensitive(False)
