@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright 2008 Zuza Software Foundation
+# Copyright 2008-2009 Zuza Software Foundation
 #
 # This file is part of Virtaal.
 #
@@ -23,14 +23,12 @@ import gtk
 import logging
 import pango
 from gtk import gdk
+from translate.lang import factory
 
 from virtaal.support.simplegeneric import generic
 from virtaal.views import markup, rendering
 from virtaal.common import pan_app
 
-import label_expander
-
-from translate.lang import factory
 
 @generic
 def compute_optimal_height(widget, width):
@@ -57,7 +55,7 @@ def make_pango_layout(widget, text, width):
     pango_layout = pango.Layout(widget.get_pango_context())
     pango_layout.set_width(width * pango.SCALE)
     pango_layout.set_wrap(pango.WRAP_WORD_CHAR)
-    pango_layout.set_text(text or "")
+    pango_layout.set_text(text or u"")
     return pango_layout
 
 @compute_optimal_height.when_type(gtk.TextView)
@@ -72,7 +70,7 @@ def gtk_textview_compute_optimal_height(widget, width):
 
     buftext = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
     if not buftext:
-        text = getattr(widget, '_source_text', "")
+        text = getattr(widget, '_source_text', u"")
         if text:
             lang = factory.getlanguage(pan_app.settings.language["targetlang"])
             buftext = lang.alter_length(text)
@@ -84,17 +82,18 @@ def gtk_textview_compute_optimal_height(widget, width):
         # directly after the file is opened. For now we try to guess a more
         # useful default than 0. This should look much better than 0, at least.
         h = 28
+    parent = widget.parent
+    if isinstance(parent, gtk.ScrolledWindow) and parent.get_shadow_type() != gtk.SHADOW_NONE:
+        border += 2 * parent.rc_get_style().ythickness
     widget.parent.set_size_request(-1, h + border)
 
-@compute_optimal_height.when_type(label_expander.LabelExpander)
-def gtk_labelexpander_compute_optimal_height(widget, width):
-    if widget.label.child.get_text().strip() == "":
-        widget.set_size_request(-1, 0)
+@compute_optimal_height.when_type(gtk.Label)
+def gtk_label_compute_optimal_height(widget, width):
+    if widget.get_text().strip() == "":
+        widget.set_size_request(width, 0)
     else:
-#        _w, h = make_pango_layout(widget, widget.label.child.get_label()[0], width).get_pixel_size()
-#        _w, h = make_pango_layout(widget, widget.label.child.get_label(), width).get_pixel_size()
-        _w, h = widget.label.child.get_layout().get_pixel_size()
-        widget.set_size_request(-1, h + 4)
+        _w, h = make_pango_layout(widget, widget.get_label(), width).get_pixel_size()
+        widget.set_size_request(width, h)
 
 
 COLUMN_NOTE, COLUMN_UNIT, COLUMN_EDITABLE = 0, 1, 2
@@ -211,7 +210,6 @@ class StoreTreeView(gtk.TreeView):
         self._enable_tooltips()
 
         self._install_callbacks()
-        self._add_accelerator_bindings()
 
         # This must be changed to a mutex if you ever consider
         # writing multi-threaded code. However, the motivation
@@ -220,26 +218,6 @@ class StoreTreeView(gtk.TreeView):
         # to you.
         self._waiting_for_row_change = 0
 
-    def _add_accelerator_bindings(self):
-        gtk.accel_map_add_entry("<Virtaal>/Navigation/Up", gtk.accelerator_parse("Up")[0], gdk.CONTROL_MASK)
-        gtk.accel_map_add_entry("<Virtaal>/Navigation/Down", gtk.accelerator_parse("Down")[0], gdk.CONTROL_MASK)
-        gtk.accel_map_add_entry("<Virtaal>/Navigation/PgUp", gtk.accelerator_parse("Page_Up")[0], gdk.CONTROL_MASK)
-        gtk.accel_map_add_entry("<Virtaal>/Navigation/PgDown", gtk.accelerator_parse("Page_Down")[0], gdk.CONTROL_MASK)
-
-        self.accel_group = gtk.AccelGroup()
-        self.accel_group.connect_by_path("<Virtaal>/Navigation/Up", self._move_up)
-        self.accel_group.connect_by_path("<Virtaal>/Navigation/Down", self._move_down)
-        self.accel_group.connect_by_path("<Virtaal>/Navigation/PgUp", self._move_pgup)
-        self.accel_group.connect_by_path("<Virtaal>/Navigation/PgDown", self._move_pgdown)
-
-        mainview = self.view.controller.main_controller.view
-        mainview.add_accel_group(self.accel_group)
-        mainview.gui.get_widget('menu_navigation').set_accel_group(self.accel_group)
-        mainview.gui.get_widget('mnu_up').set_accel_path('<Virtaal>/Navigation/Up')
-        mainview.gui.get_widget('mnu_down').set_accel_path('<Virtaal>/Navigation/Down')
-        mainview.gui.get_widget('mnu_pageup').set_accel_path('<Virtaal>/Navigation/PgUp')
-        mainview.gui.get_widget('mnu_pagedown').set_accel_path('<Virtaal>/Navigation/PgDown')
-
     def _enable_tooltips(self):
         if hasattr(self, "set_tooltip_column"):
             self.set_tooltip_column(COLUMN_NOTE)
@@ -301,6 +279,9 @@ class StoreTreeView(gtk.TreeView):
         gobject.idle_add(change_cursor, priority=gobject.PRIORITY_DEFAULT_IDLE)
 
     def _keyboard_move(self, offset):
+        if not self.view.controller.get_store():
+            return
+
         # We don't want to process keyboard move events until we have finished updating
         # the display after a move event. So we use this awful, awful, terrible scheme to
         # keep track of pending draw events. In reality, it should be impossible for
@@ -480,12 +461,16 @@ class StoreCellRenderer(gtk.GenericCellRenderer):
             editor.set_size_request(width, -1)
             editor.show()
             compute_optimal_height(editor, width)
+            parent_height = widget.get_allocation().height
+            if parent_height < -1:
+                parent_height = widget.size_request()[1]
+            if parent_height > 0:
+                self.check_editor_height(editor, width, parent_height)
             _width, height = editor.size_request()
-            #We used to have a minimum height. Do we want this?
-            #height = max(height, 70)
+            height += self.ROW_PADDING
         else:
             height = self.compute_cell_height(widget, width)
-        height = min(height, 600)
+        #height = min(height, 600)
         y_offset = self.ROW_PADDING / 2
         return 0, y_offset, width, height
 
@@ -512,7 +497,7 @@ class StoreCellRenderer(gtk.GenericCellRenderer):
         if widget.get_direction() == gtk.TEXT_DIR_LTR:
             target_x += width/2
         else:
-            source_x += width/2
+            source_x += (width/2) + 10
         widget.get_style().paint_layout(window, gtk.STATE_NORMAL, False,
                 cell_area, widget, '', source_x, y, self.source_layout)
         widget.get_style().paint_layout(window, gtk.STATE_NORMAL, False,
@@ -530,13 +515,14 @@ class StoreCellRenderer(gtk.GenericCellRenderer):
         layout.set_wrap(pango.WRAP_WORD_CHAR)
         layout.set_width(width * pango.SCALE)
         #XXX - plurals?
-        text = text or ""
+        text = text or u""
         layout.set_markup(markup.markuptext(text))
         return layout
 
     def compute_cell_height(self, widget, width):
-        srclang = self.view.controller.main_controller.lang_controller.source_lang.code
-        tgtlang = self.view.controller.main_controller.lang_controller.target_lang.code
+        lang_controller = self.view.controller.main_controller.lang_controller
+        srclang = lang_controller.source_lang.code
+        tgtlang = lang_controller.target_lang.code
         self.source_layout = self._get_pango_layout(widget, self.unit.source, width / 2,
                 rendering.get_source_font_description())
         self.source_layout.get_context().set_language(rendering.get_language(srclang))
@@ -548,10 +534,34 @@ class StoreCellRenderer(gtk.GenericCellRenderer):
         if widget.get_direction() == gtk.TEXT_DIR_RTL:
             self.source_layout.set_alignment(pango.ALIGN_RIGHT)
             self.target_layout.set_alignment(pango.ALIGN_RIGHT)
+            self.target_layout.set_auto_dir(False)
         _layout_width, source_height = self.source_layout.get_pixel_size()
         _layout_width, target_height = self.target_layout.get_pixel_size()
         return max(source_height, target_height) + self.ROW_PADDING
 
+    def check_editor_height(self, editor, width, parentheight):
+        notesheight = 0
+
+        for note in editor._widgets['notes'].values():
+            notesheight += note.size_request()[1]
+
+        maxheight = parentheight - notesheight
+
+        if maxheight < 0:
+            return
+
+        visible_textboxes = []
+        for textbox in (editor._widgets['sources'] + editor._widgets['targets']):
+            if textbox.props.visible:
+                visible_textboxes.append(textbox)
+
+        max_tb_height = maxheight / len(visible_textboxes)
+
+        for textbox in visible_textboxes:
+            if textbox.props.visible and textbox.parent.size_request()[1] > max_tb_height:
+                textbox.parent.set_size_request(-1, max_tb_height)
+                #logging.debug('%s.set_size_request(-1, %d)' % (textbox.parent, max_tb_height))
+
 
     # EVENT HANDLERS #
     def _on_editor_done(self, editor):
